-------------------------------------------------------------------------------------
-- Title: Mik's Scrolling Battle Text
-- Author: Mik
-------------------------------------------------------------------------------------

-------------------------------------------------------------------------------------
-- Public Constants.
-------------------------------------------------------------------------------------

-- Display types for incoming, outgoing, and notification events.
MikSBT.DISPLAYTYPE_INCOMING			= 1;
MikSBT.DISPLAYTYPE_OUTGOING			= 2;
MikSBT.DISPLAYTYPE_NOTIFICATION		= 3;


-------------------------------------------------------------------------------------
-- Private constants.
-------------------------------------------------------------------------------------

-- Number of font strings.
local NUM_FONT_STRINGS = 15;

-- Animation values for incoming and outgoing events.
local MIN_VERTICAL_SPACING = 8;

-- How long to wait before showing events so that merges may happen.
local MERGE_DELAY_TIME = 0.3;

-- How long to show stickies before they are faded.
local STICKY_FADE_IN_TIME = 0.15;
local STICKY_DISPLAY_TIME = 2;
local STICKY_FADE_OUT_TIME = 0.5;

-- The animation speed.
local ANIMATION_SPEED = 0.015;

-- Scroll directions
local SCROLL_DIRECTION_UP			= 1;
local SCROLL_DIRECTION_DOWN			= 2;

-- Animation Styles
local ANIMATION_STYLE_STRAIGHT		= 1;
local ANIMATION_STYLE_CURVED_LEFT		= 2;
local ANIMATION_STYLE_CURVED_RIGHT		= 3;

-- Max number of combo points.
local MAX_COMBO_POINTS = 5;


-------------------------------------------------------------------------------------
-- Private variables.
-------------------------------------------------------------------------------------

-- Holds whether or not debugging is active.
local debugMode = false;


-- Holds the events the mod is interested in.
local listenEvents = {};

-- Holds the current profile.
local currentProfile;

-- Holds data for each scroll area.
local scrollAreas = {};

-- Holds arrays of unmerged events, merged events to be displayed and the time since
-- the last merge for incoming and outgoing events.
local animationMergeData = {
 ["Incoming"]	= {["TimeSinceMerge"] = 0, UnmergedEvents={}, MergedEvents={}},
 ["Outgoing"]	= {["TimeSinceMerge"] = 0, UnmergedEvents={}, MergedEvents={}},
};

-- Holds whether or not a merge is still being processed.
local stillMerging = false;

-- Temp arrays for repositioning animation display data to avoid overlaps.
local activeNonStickies = {};
local activeStickies = {};

-- Create a new table recycler object that holds up to 50 table references
-- in primary cache.
local eventsRecycler = MikTRO:NewRecyclerObject(50);


-------------------------------------------------------------------------------------
-- Printing functions.
-------------------------------------------------------------------------------------

-- **********************************************************************************
-- Prints out the passed message to the default chat frame.
-- **********************************************************************************
function MikSBT.Print(msg, r, g, b)
 -- Add the message to the default chat frame.
 DEFAULT_CHAT_FRAME:AddMessage("MSBT: " .. msg, r, g, b);
end


-- **********************************************************************************
-- Prints out the passed message to the default chat frame.
-- **********************************************************************************
function MikSBT.PrintDebug(msg, r, g, b)
 -- Check if debugging is active.
 if (debugMode) then
  -- Add the message to the default chat frame.
  DEFAULT_CHAT_FRAME:AddMessage(msg, r, g, b);
 end
end


-------------------------------------------------------------------------------------
-- Core event handlers.
-------------------------------------------------------------------------------------

-- **********************************************************************************
-- Registers all of the core events the mod is interested in.
-- **********************************************************************************
function MikSBT.RegisterEvents()
 -- Register the events we are interested in receiving for shaman.
 for k, v in listenEvents do
  MSBTEventFrame:RegisterEvent(v);
 end
end


-- **********************************************************************************
-- Unregisters all of the events the mod registered for.
-- **********************************************************************************
function MikSBT.UnregisterEvents()
 for k, v in listenEvents do
  MSBTEventFrame:UnregisterEvent(v);
 end
end


-- **********************************************************************************
-- Called when the mod is loaded.
-- **********************************************************************************
function MikSBT.OnLoad()
 -- Register for the ADDON_LOADED event.
 MSBTEventFrame:RegisterEvent("ADDON_LOADED");
end

-- **********************************************************************************
-- Called when the events we registered for occur.
-- **********************************************************************************
function MikSBT.OnEvent()
 -- When an addon is loaded.
 if (event == "ADDON_LOADED") then
  -- Make sure it's this addon.
  if (arg1 == MikSBT.MOD_NAME) then

   -- Don't get notification for other addons being loaded.
   this:UnregisterEvent("ADDON_LOADED");

   -- Register for the events the helper is interested in receiving.
   MikSBT.RegisterEvents();

   -- Initialize the mod.
   MikSBT.Init();
  end

 end
end


-- **********************************************************************************
-- Called when the core events frame is updated.
-- **********************************************************************************
function MikSBT.OnUpdate()

 -- Loop through all of the merge data tables.
 for _, mergeData in animationMergeData do
  -- Add the elapsed time since the last merge.
  mergeData.TimeSinceMerge = mergeData.TimeSinceMerge + arg1;

  -- Once enough time has passed and all previous merges are done.
  if ((mergeData.TimeSinceMerge >= MERGE_DELAY_TIME) and not stillMerging) then
   -- Set the still merging flag.
   stillMerging = true;

   -- Merge like events.
   local numEvents = table.getn(mergeData.UnmergedEvents);
   MikSBT.MergeEvents(mergeData, numEvents);

   -- Add the merged animation events to the animation system.
   for _, animationEvent in mergeData.MergedEvents do
    MikSBT.AddAnimation(animationEvent);
   end

   -- Clear the merged animation events array.
   local numMergedEvents = table.getn(mergeData.MergedEvents);
   for i = 1, numMergedEvents do
    table.remove(mergeData.MergedEvents, 1);
   end


   -- Reset the last merged time.
   mergeData.TimeSinceMerge = 0;

   -- Clear the still merging flag.
   stillMerging = false;
  end
 end
end


-- **********************************************************************************
-- Called when the mod's animation frames are updated.
-- **********************************************************************************
function MikSBT.OnUpdateAnimationFrame(frame)
 local scrollArea;

 -- Get the frame's name.
 local frameName = frame:GetName();
 if (frameName == scrollAreas.Incoming.AssociatedFrame) then
  scrollArea = scrollAreas.Incoming;
 elseif (frameName == scrollAreas.Outgoing.AssociatedFrame) then
  scrollArea = scrollAreas.Outgoing;
 elseif (frameName == scrollAreas.Notification.AssociatedFrame) then
  scrollArea = scrollAreas.Notification;
 end

 -- Flag for whether or not all the animations are inactive.
 local allInactive = true;

 -- Loop through all possible animation display objects for the scroll area.
 for _, animDisplayInfo in scrollArea.AnimationDisplayArray do
  -- Check if the animation is active.
  if (animDisplayInfo.IsActive) then
   -- Do the animation and clear the all inactive flag.
   MikSBT.DoAnimation(animDisplayInfo, scrollArea);
   allInactive = false;
  end
 end

 -- Check if none of the animations are active and the events frame is visible.  If so, hide it so
 -- the OnUpdate events stop firing.  This is done to keep the number of OnUpdate events down to a minimum for
 -- better performance.
 if (allInactive and frame:IsVisible()) then
  frame:Hide();
 end
end


-- **********************************************************************************
-- Called when the mod is fully loaded (variables loaded).
-- **********************************************************************************
function MikSBT.Init()
 -- Register slash commands
 SLASH_MSBT1 = MikSBT.COMMAND;
 SlashCmdList["MSBT"] = function (params)
  MikSBT.CommandHandler(params);
 end

 -- Check if there are no saved variables.
 if (not MikSBT_Save or not MikSBT_Save.Profiles) then
  MikSBT_Save = {};

  MikSBT_Save.UserDisabled = false;
  MikSBT_Save.CurrentProfile = MikSBT.DEFAULT_PROFILE_NAME;
  MikSBT_Save.Profiles = {};
  MikSBT_Save.Profiles[MikSBT_Save.CurrentProfile] = {};

  -- Set all the options to their defaults.
  MikSBT.ResetProfile(MikSBT_Save.CurrentProfile);
 end

 -- Updates profiles created by older versions.
 MikSBT.UpdateProfiles();


 -- Set the current profile.
 currentProfile = MikSBT_Save.Profiles[MikSBT_Save.CurrentProfile];

 -- Sets up all of the animation scroll areas.
 MikSBT.SetupAnimationScrollAreas();

 -- Reposition the animation frames to the stored offsets.
 MikSBT.RepositionAnimationFrames();

 -- Disable or enable the mod depending on the saved setting.
 MikSBT.SetOptionUserDisabled(MikSBT_Save.UserDisabled)
end


-- **********************************************************************************
-- Sets up all of the animation scroll areas.
-- **********************************************************************************
function MikSBT.SetupAnimationScrollAreas()
 -- Setup a table for data for each scroll area.
 scrollAreas.Incoming = {};
 scrollAreas.Outgoing = {};
 scrollAreas.Notification = {};

 -- Setup the associated frames.
 scrollAreas.Incoming.AssociatedFrame = "MSBTFrameIncoming";
 scrollAreas.Outgoing.AssociatedFrame = "MSBTFrameOutgoing";
 scrollAreas.Notification.AssociatedFrame = "MSBTFrameNotification";

 -- Setup the display settings.
 scrollAreas.Incoming.DisplaySettings = currentProfile.DisplaySettings.Incoming;
 scrollAreas.Outgoing.DisplaySettings = currentProfile.DisplaySettings.Outgoing;
 scrollAreas.Notification.DisplaySettings = currentProfile.DisplaySettings.Notification;


 -- Loop through the scroll areas.
 for _, scrollArea in scrollAreas do
  -- Initialize the next available animation display object.
  scrollArea.NextAnimationDisplay = 1;

  -- Initialize the next available sticky animation display object.
  scrollArea.FirstSticky = 11;
  scrollArea.NextSticky = scrollArea.FirstSticky;

  -- Create arrays for the animation display info.
  scrollArea.AnimationDisplayArray = {};

  -- Associate the appropriate font string object for each animation display object.
  for x = 1, NUM_FONT_STRINGS do
   scrollArea.AnimationDisplayArray[x] = {};
   scrollArea.AnimationDisplayArray[x].FontObject = getglobal(scrollArea.AssociatedFrame .. "Text" .. x);
   scrollArea.AnimationDisplayArray[x].LastUpdated = 0;
  end
 end
end


-- **********************************************************************************
-- Repositions the animation frames based on the stored offsets.
-- **********************************************************************************
function MikSBT.RepositionAnimationFrames()
 -- Loop through each scroll area and reposition its associated frame.
 for _, scrollArea in scrollAreas do
  getglobal(scrollArea.AssociatedFrame):SetPoint("BOTTOM", "UIParent", "CENTER", scrollArea.DisplaySettings.FrameOffsets.x, scrollArea.DisplaySettings.FrameOffsets.y);
 end
end


-- **********************************************************************************
-- Receive combat events from Mik's combat event helper.
-- **********************************************************************************
function MikSBT.CombatEventsNotificationFunction(combatEvent)
 -- Get a recycled table from the events recycler object.
 local animationEvent = eventsRecycler:AcquireTable();

 -- Initalize the event type string to the common prefix.
 local eventTypeString = "MSBT_EVENTTYPE_";

 -- Append the appropriate direction type to the event type string.
 if (combatEvent.DirectionType == MikCEH.DIRECTIONTYPE_PLAYER_INCOMING) then
  animationEvent.DisplayType = MikSBT.DISPLAYTYPE_INCOMING;
  eventTypeString = eventTypeString .. "INCOMING_";
 elseif (combatEvent.DirectionType == MikCEH.DIRECTIONTYPE_PLAYER_OUTGOING) then
  animationEvent.DisplayType = MikSBT.DISPLAYTYPE_OUTGOING;
  eventTypeString = eventTypeString .. "OUTGOING_";
 elseif (combatEvent.DirectionType == MikCEH.DIRECTIONTYPE_PET_OUTGOING) then
  animationEvent.DisplayType = MikSBT.DISPLAYTYPE_OUTGOING;
  eventTypeString = eventTypeString .. "OUTGOING_PET_";
 end

 -- Check if it's a damage event.
 if (combatEvent.EventType == MikCEH.EVENTTYPE_DAMAGE) then
  -- Check if there is an associated spell/ability name.
  if (combatEvent.EffectName ~= nil) then
   eventTypeString = eventTypeString .. "SPELL_";
  end

  -- Append the appropriate action type to the event type string.
  if (combatEvent.ActionType == MikCEH.ACTIONTYPE_HIT) then
   if (combatEvent.HitType == MikCEH.HITTYPE_OVER_TIME) then
    eventTypeString = eventTypeString .. "DOT";
   else
    eventTypeString = eventTypeString .. "DAMAGE";
   end
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_MISS) then
   eventTypeString = eventTypeString .. "MISS";
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_DODGE) then
   eventTypeString = eventTypeString .. "DODGE";
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_PARRY) then
   eventTypeString = eventTypeString .. "PARRY";
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_BLOCK) then
   eventTypeString = eventTypeString .. "BLOCK";
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_RESIST) then
   eventTypeString = eventTypeString .. "RESIST";
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_ABSORB) then
   eventTypeString = eventTypeString .. "ABSORB";
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_IMMUNE) then
   eventTypeString = eventTypeString .. "IMMUNE";
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_EVADE) then
   eventTypeString = eventTypeString .. "EVADE";
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_REFLECT) then
   eventTypeString = eventTypeString .. "REFLECT";
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_DROWNING) then
   eventTypeString = eventTypeString .. "ENVIRONMENTAL"
   combatEvent.EffectName = MikSBT.MSG_ENVIRONMENTAL_DROWNING;
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_FALLING) then
   eventTypeString = eventTypeString .. "ENVIRONMENTAL"
   combatEvent.EffectName = MikSBT.MSG_ENVIRONMENTAL_FALLING;
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_FATIGUE) then
   eventTypeString = eventTypeString .. "ENVIRONMENTAL"
   combatEvent.EffectName = MikSBT.MSG_ENVIRONMENTAL_FATIGUE;
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_FIRE) then
   eventTypeString = eventTypeString .. "ENVIRONMENTAL"
   combatEvent.EffectName = MikSBT.MSG_ENVIRONMENTAL_FIRE;
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_LAVA) then
   eventTypeString = eventTypeString .. "ENVIRONMENTAL"
   combatEvent.EffectName = MikSBT.MSG_ENVIRONMENTAL_LAVA;
  elseif (combatEvent.ActionType == MikCEH.ACTIONTYPE_SLIME) then
   eventTypeString = eventTypeString .. "ENVIRONMENTAL"
   combatEvent.EffectName = MikSBT.MSG_ENVIRONMENTAL_SLIME;
  end

 -- Heal event.
 elseif (combatEvent.EventType == MikCEH.EVENTTYPE_HEAL) then
   if (combatEvent.HealType == MikCEH.HEALTYPE_NORMAL or combatEvent.HealType == MikCEH.HEALTYPE_CRIT) then
     eventTypeString = eventTypeString .. "HEAL";
   elseif (combatEvent.HealType == MikCEH.HEALTYPE_OVER_TIME) then
    eventTypeString = eventTypeString .. "HOT";
   end
    

 -- Notification event.
 elseif (combatEvent.EventType == MikCEH.EVENTTYPE_NOTIFICATION) then
  animationEvent.DisplayType = MikSBT.DISPLAYTYPE_NOTIFICATION;
  eventTypeString = eventTypeString .. "NOTIFICATION_";

  -- Append the appropriate notification type to the event type string.
  if (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_DEBUFF) then
   eventTypeString = eventTypeString .. "DEBUFF";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_BUFF) then
   eventTypeString = eventTypeString .. "BUFF";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_ITEM_BUFF) then
   eventTypeString = eventTypeString .. "ITEM_BUFF";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_BUFF_FADE) then
   eventTypeString = eventTypeString .. "BUFF_FADE";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_COMBAT_ENTER) then
   eventTypeString = eventTypeString .. "COMBAT_ENTER";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_COMBAT_LEAVE) then
   eventTypeString = eventTypeString .. "COMBAT_LEAVE";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_POWER_GAIN) then
   eventTypeString = eventTypeString .. "POWER_GAIN";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_POWER_LOSS) then
   eventTypeString = eventTypeString .. "POWER_LOSS";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_CP_GAIN) then
   -- Check if we are at max combo points.
   if (combatEvent.Amount == MAX_COMBO_POINTS) then
    eventTypeString = eventTypeString .. "CP_FULL";
   else
    eventTypeString = eventTypeString .. "CP_GAIN";
   end
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_HONOR_GAIN) then
   eventTypeString = eventTypeString .. "HONOR_GAIN";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_REP_GAIN) then
   eventTypeString = eventTypeString .. "REP_GAIN";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_REP_LOSS) then
   eventTypeString = eventTypeString .. "REP_LOSS";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_SKILL_GAIN) then
   eventTypeString = eventTypeString .. "SKILL_GAIN";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_EXPERIENCE_GAIN) then
   eventTypeString = eventTypeString .. "EXPERIENCE_GAIN";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_LOW_HEALTH) then
   eventTypeString = eventTypeString .. "LOW_HEALTH";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_LOW_MANA) then
   eventTypeString = eventTypeString .. "LOW_MANA";
  elseif (combatEvent.NotificationType == MikCEH.NOTIFICATIONTYPE_EXECUTE) then
   eventTypeString = eventTypeString .. "EXECUTE";
  elseif (combatEvent.NotificationType ~= nil) then
   eventTypeString = eventTypeString .. combatEvent.NotificationType;
  end
 end

 -- Check if the event is a crit and set the crit field appropriately
 if (combatEvent.HitType == MikCEH.HITTYPE_CRIT or combatEvent.HealType == MikCEH.HEALTYPE_CRIT) then
  animationEvent.IsCrit = true;
 end

 -- Copy the amount, effect name, and player/enemy name from the passed combat event data.
 animationEvent.Amount = combatEvent.Amount;
 animationEvent.EffectName = combatEvent.EffectName;
 animationEvent.Name = combatEvent.Name;

 -- Check if there is a damage type and get the appropriate string.
 if (combatEvent.DamageType ~= nil) then
  animationEvent.DamageType = MikCEH.GetDamageTypeString(combatEvent.DamageType);
 end

 -- If there are partial action types set the partial effect text appropriately.
 if (combatEvent.PartialActionType == MikCEH.PARTIALACTIONTYPE_ABSORB) then
  if (ABSORB_TRAILER) then
   animationEvent.PartialEffectText = string.gsub(ABSORB_TRAILER, "%%d", combatEvent.PartialAmount);
  end
 elseif (combatEvent.PartialActionType == MikCEH.PARTIALACTIONTYPE_BLOCK) then
  if (BLOCK_TRAILER) then
   animationEvent.PartialEffectText = string.gsub(BLOCK_TRAILER, "%%d", combatEvent.PartialAmount);
  end
 elseif (combatEvent.PartialActionType == MikCEH.PARTIALACTIONTYPE_RESIST) then
  if (RESIST_TRAILER) then
   animationEvent.PartialEffectText = string.gsub(RESIST_TRAILER, "%%d", combatEvent.PartialAmount);
  end
 elseif (combatEvent.PartialActionType == MikCEH.PARTIALACTIONTYPE_VULNERABLE) then
  if (MikSBT.MSG_VULERNABLE_TRAILER) then
   animationEvent.PartialEffectText = string.gsub(MikSBT.MSG_VULERNABLE_TRAILER, "%%d", combatEvent.PartialAmount);
  end
 elseif (combatEvent.PartialActionType == MikCEH.PARTIALACTIONTYPE_OVERHEAL) then
  animationEvent.OverhealAmount = combatEvent.PartialAmount;
 end



 -- Check if there is profile data for the event type.
 if (currentProfile.EventSettings[eventTypeString]) then
  -- Set the event type, its unformatted text and dispatch it to the animation system.
  animationEvent.EventType = eventTypeString;
  animationEvent.UnformattedText = currentProfile.EventSettings[eventTypeString].Message;

  -- Dispatch the event to the animation system.
  MikSBT.DispatchAnimationEvent(animationEvent);
 else
  MikSBT.Print("No profile data found for " .. eventTypeString, 1, 0, 0)
 end
end


-------------------------------------------------------------------------------------
-- Utility functions.
-------------------------------------------------------------------------------------

-- **********************************************************************************
-- Displays the passed message in the passed scroll area with the passed parameters.
-- This function is for easy displaying of messages from external sources.
-- **********************************************************************************
function MikSBT.DisplayMessage(message, displayType, isSticky, colorR, colorG, colorB, fontSize, fontIndex, outlineIndex)
 -- Check if no message was passed and do nothing.
 if (message == nil) then
  return;
 end

 -- Check if the passed display type is invalid..
 if (displayType ~= MikSBT.DISPLAYTYPE_INCOMING and displayType ~= MikSBT.DISPLAYTYPE_OUTGOING and
     displayType ~= MikSBT.DISPLAYTYPE_NOTIFICATION) then

  -- Set the display type to the default.
  displayType = MikSBT.DISPLAYTYPE_NOTIFICATION;
 end

 -- Holds the correct scroll area for the event.
 local scrollArea;

 -- Get the correct scroll area for the event.
 if (displayType == MikSBT.DISPLAYTYPE_INCOMING) then
  scrollArea = scrollAreas.Incoming;  
 elseif (displayType == MikSBT.DISPLAYTYPE_OUTGOING) then
  scrollArea = scrollAreas.Outgoing;
 elseif (displayType == MikSBT.DISPLAYTYPE_NOTIFICATION) then
  scrollArea = scrollAreas.Notification;
 end

 -- Check if the red color value is invalid.
 if (colorR == nil or colorR < 0 or colorR > 255) then
  -- Set the red color value to the default.
  colorR = 255;
 end

 -- Check if the green color value is invalid.
 if (colorG == nil or colorG < 0 or  colorG > 255) then
  -- Set the green color value to the default.
  colorG = 255;
 end

 -- Check if the blue color value is invalid.
 if (colorB == nil or colorB < 0 or  colorB > 255) then
  -- Set the blue color value to the default.
  colorB = 255;
 end

 -- Check if the font size is invalid.
 if (fontSize == nil or fontSize < 12 or fontSize > 32) then
  fontSize = MikSBT.GetFontSize(scrollArea, nil, false);
 end

 local fontPath;
 -- Check if the font index in invalid.
 if (fontIndex == nil or MikSBT.AVAILABLE_FONTS[fontIndex] == nil) then
  fontPath = MikSBT.GetFontPath(scrollArea, nil, false);
 -- Font index is valid.
 else
  fontPath = MikSBT.AVAILABLE_FONTS[fontIndex].Path;
 end

 local fontOutline;
 -- Check if the outline index in invalid.
 if (outlineIndex == nil or MikSBT.AVAILABLE_OUTLINES[outlineIndex] == nil) then
  fontOutline = MikSBT.GetFontOutline(scrollArea, nil, false);
 else
  fontOutline = MikSBT.AVAILABLE_OUTLINES[outlineIndex].Style;
 end


 -- Get a recycled table from the events recycler object.
 local animationEvent = eventsRecycler:AcquireTable();

 -- Populate the animation event data.
 animationEvent.DisplayType = displayType;
 animationEvent.IsCrit = false;
 animationEvent.IsSticky = isSticky;
 animationEvent.Text = message;
 animationEvent.FontSize = fontSize;
 animationEvent.FontPath = fontPath;
 animationEvent.FontOutline = fontOutline;
 animationEvent.ColorR = colorR;
 animationEvent.ColorG = colorG;
 animationEvent.ColorB = colorB;


 -- Add the event to the animation system.
 MikSBT.AddAnimation(animationEvent);
end


-- **********************************************************************************
-- This function copies a table.
-- **********************************************************************************
function MikSBT.CopyTable(srcTable)
 -- Create a new table.
 local newTable = {};

 -- Loop through all of the entires in the table.
 for key, value in srcTable do
  -- Check if it's a nested table.
  if (type(value) == "table") then
   -- Recursively call the function to copy the nested table.
   value = MikSBT.CopyTable(value);
  end

  -- Make a copy of the value into the new table.
  newTable[key] = value;
 end

 -- Return the new table.
 return newTable;
end


-- **********************************************************************************
-- Merges like animation events.
-- **********************************************************************************
function MikSBT.MergeEvents(mergeData, numEvents)
 -- Hold whether or not the event was merged.
 local eventMerged = false;

 -- Don't attempt to merge any more events than were available when the function was called since
 -- more events may get added while the merge is taking place.
 for x = 1, numEvents do

  -- Loop through all of the events in the merged events array.
  for _, animationEvent in mergeData.MergedEvents do
   -- Check if the event types match and the effect names are the same.
   if ((mergeData.UnmergedEvents[x].EventType == animationEvent.EventType) and
       (mergeData.UnmergedEvents[x].EffectName == animationEvent.EffectName)) then

    -- Check if there is no effect name
    if (mergeData.UnmergedEvents[x].EffectName == nil) then
     -- Check if the affected unit name is the same.
     if ((mergeData.UnmergedEvents[x].Name == animationEvent.Name) and mergeData.UnmergedEvents[x].Name ~= nil) then
      -- Do Merge.
      eventMerged = true;
     end -- Same name.

    -- There is an effect name.
    else
     -- Check if there is a name and replace it with the multiple targets string.
     if (animationEvent.Name ~= nil) then
      animationEvent.Name = MikSBT.MSG_MULTIPLE_TARGETS;
     end

     -- Do Merge.
     eventMerged = true;
    end -- Effect name?
   end -- Event types match and effect names the same.


   -- Check if the event should be merged.
   if (eventMerged) then
    -- Set the event merged flag for the event being merged.
    mergeData.UnmergedEvents[x].EventMerged = true;

    -- If the events have an amount then total them.
    if (animationEvent.Amount ~= nil and mergeData.UnmergedEvents[x].Amount ~= nil) then
     animationEvent.Amount = animationEvent.Amount + mergeData.UnmergedEvents[x].Amount;
    end

    -- Check if there is an overheal amount and total it.
    if (mergeData.UnmergedEvents[x].OverhealAmount ~= nil) then
     animationEvent.OverhealAmount = (animationEvent.OverhealAmount or 0) + mergeData.UnmergedEvents[x].OverhealAmount;
    end

    -- Clear any partial effect text.
    animationEvent.PartialEffectText = nil;

    
    -- Increment the number of merged events.
    animationEvent.NumMerged = animationEvent.NumMerged + 1;

    -- Check if the event being merged is a crit and increment the number of crits.
    if (mergeData.UnmergedEvents[x].IsCrit) then
     animationEvent.NumCrits = animationEvent.NumCrits + 1;
    end

    -- Make sure the merged event is not flagged as a crit.
    animationEvent.IsCrit = false;
   end -- Do Merge.
  end -- Loop through all merged events.


  -- If the event wasn't merged add it to the end of the merged events array.
  if (not eventMerged) then
   local animationEvent = mergeData.UnmergedEvents[x];
   animationEvent.NumMerged = 0;

   -- Check if the event is a crit and set the number of crits appropriately.
   if (animationEvent.IsCrit) then
    animationEvent.NumCrits = 1;
   else
    animationEvent.NumCrits = 0;
   end

   -- Add the animation event to the end of the merged events array.
   table.insert(mergeData.MergedEvents, animationEvent);
  end

  -- Reset the event merged flag.
  eventMerged = false;
 end


 -- Loop through each event in the merged events array.
 for _, animationEvent in mergeData.MergedEvents do
  -- Check if there were any events merged.
  if (animationEvent.NumMerged > 0) then
   -- Create trailer text with the number of merged events.
   local trailerText = " [" .. animationEvent.NumMerged + 1 .. " " .. MikSBT.MSG_HITS;

   -- Check if there were any crits merged.
   if (animationEvent.NumCrits > 1) then
    trailerText = trailerText .. ", " .. animationEvent.NumCrits .. " " .. MikSBT.MSG_CRITS;
   elseif (animationEvent.NumCrits == 1) then
    trailerText = trailerText .. ", " .. animationEvent.NumCrits .. " " .. MikSBT.MSG_CRIT;
   end

   -- Close the merge string.
   trailerText = trailerText .. "]";

   -- Set the animation event's merge trailer text field.
   animationEvent.MergeTrailerText = trailerText;
  end
 end


 -- Remove the processed animation events from unmerged events queue.
 for x = 1, numEvents do
  -- Check if the event was merged into a different animation event.
  if (mergeData.UnmergedEvents[1].EventMerged) then
   -- Reclaim the animation event table to the events recycler.
   eventsRecycler:ReclaimTable(mergeData.UnmergedEvents[1]);
  end

  -- Remove the event from the unmerged events array.
  table.remove(mergeData.UnmergedEvents, 1);
 end
end


-- **********************************************************************************
-- This function gets the appropriate font size for an event.
-- **********************************************************************************
function MikSBT.GetFontSize(scrollArea, eventType, isCrit)
 -- Holds the font sizes for the event level.
 local eventFontSizeNormal = 0;
 local eventFontSizeCrit = 0;

 -- Check if there is an event type.
 if (eventType ~= nil) then
  -- Get the font sizes for the event type.
  eventFontSizeNormal = currentProfile.EventSettings[eventType].FontSettings.Normal.FontSize;

  -- Check if there are crit font settings for the event.
  if (currentProfile.EventSettings[eventType].FontSettings.Crit ~= nil) then
   eventFontSizeCrit = currentProfile.EventSettings[eventType].FontSettings.Crit.FontSize;
  end
 end

  -- Get the font sizes for the scroll area and master.
 local scrollAreaFontSizeNormal = scrollArea.DisplaySettings.FontSettings.Normal.FontSize;
 local scrollAreaFontSizeCrit = scrollArea.DisplaySettings.FontSettings.Crit.FontSize;
 local masterFontSizeNormal = currentProfile.MasterFontSettings.Normal.FontSize;
 local masterFontSizeCrit = currentProfile.MasterFontSettings.Crit.FontSize;


 -- Get the normal font size for the event.
 local fontSize = eventFontSizeNormal;

 -- Check if the font size is inherited from the scroll area font size.
 if (fontSize == 0) then
  fontSize = scrollAreaFontSizeNormal;
 end

 -- Check if the font size for the scroll area is inherited from the master font size.
 if (fontSize == 0) then
  fontSize = masterFontSizeNormal;
 end


 -- Check if the event is a crit.
 if (isCrit) then
  -- Get the crit font size for the event.
  fontSize = eventFontSizeCrit;

  -- Check if the font size is inherited from the scroll area font size.
  if (fontSize == 0) then
   fontSize = scrollAreaFontSizeCrit;
  end

  -- Check if the font size for the scroll area is inherited from the master font size.
  if (fontSize == 0) then
   fontSize = masterFontSizeCrit;
  end
 end

 -- Return the font size.
 return fontSize;
end


-- **********************************************************************************
-- This function gets the appropriate font path for an event.
-- **********************************************************************************
function MikSBT.GetFontPath(scrollArea, eventType, isCrit)
 -- Holds the font indices for the event level.
 local eventFontIndexNormal = 0;
 local eventFontIndexCrit = 0;

 -- Check if there is an event type.
 if (eventType ~= nil) then
  -- Get the font indices for the event type.
  eventFontIndexNormal = currentProfile.EventSettings[eventType].FontSettings.Normal.FontIndex;

  -- Check if there are crit font settings for the event.
  if (currentProfile.EventSettings[eventType].FontSettings.Crit ~= nil) then
   eventFontIndexCrit = currentProfile.EventSettings[eventType].FontSettings.Crit.FontIndex;
  end
 end

 -- Get the font indices for the scroll area and master.
 local scrollAreaFontIndexNormal = scrollArea.DisplaySettings.FontSettings.Normal.FontIndex;
 local scrollAreaFontIndexCrit = scrollArea.DisplaySettings.FontSettings.Crit.FontIndex;
 local masterFontIndexNormal = currentProfile.MasterFontSettings.Normal.FontIndex;
 local masterFontIndexCrit = currentProfile.MasterFontSettings.Crit.FontIndex;


 -- Get the normal font index for the event.
 local fontIndex = eventFontIndexNormal;

 -- Check if the font index is inherited from the scroll area font index.
 if (fontIndex == 0) then
  fontIndex = scrollAreaFontIndexNormal;
 end

 -- Check if the font index for the scroll area is inherited from the master font index.
 if (fontIndex == 0) then
  fontIndex = masterFontIndexNormal;
 end


 -- Check if the event is a crit.
 if (isCrit) then
  -- Get the crit font index for the event.
  fontIndex = eventFontIndexCrit;

  -- Check if the font index is inherited from the scroll area font index.
  if (fontIndex == 0) then
   fontIndex = scrollAreaFontIndexCrit;
  end

  -- Check if the font index for the scroll area is inherited from the master font index.
  if (fontIndex == 0) then
   fontIndex = masterFontIndexCrit;
  end
 end

 -- Check if the font index is invalid and set it to the default if it is.
 if (MikSBT.AVAILABLE_FONTS[fontIndex] == nil) then
  fontIndex = 1;
 end

 -- Return the font path.
 return MikSBT.AVAILABLE_FONTS[fontIndex].Path;
end


-- **********************************************************************************
-- This function gets the appropriate font outline style for an event.
-- **********************************************************************************
function MikSBT.GetFontOutline(scrollArea, eventType, isCrit)
 -- Holds the outline indices for the event level.
 local eventOutlineIndexNormal = 0;
 local eventOutlineIndexCrit = 0;

 -- Check if there is an event type.
 if (eventType ~= nil) then
  -- Get the outline indices for the event type.
  eventOutlineIndexNormal = currentProfile.EventSettings[eventType].FontSettings.Normal.OutlineIndex;

  -- Check if there are crit font settings for the event.
  if (currentProfile.EventSettings[eventType].FontSettings.Crit ~= nil) then
   eventOutlineIndexCrit = currentProfile.EventSettings[eventType].FontSettings.Crit.OutlineIndex;
  end
 end

 -- Get the outline indices for the scroll area and master.
 local scrollAreaOutlineIndexNormal = scrollArea.DisplaySettings.FontSettings.Normal.OutlineIndex;
 local scrollAreaOutlineIndexCrit = scrollArea.DisplaySettings.FontSettings.Crit.OutlineIndex;
 local masterOutlineIndexNormal = currentProfile.MasterFontSettings.Normal.OutlineIndex;
 local masterOutlineIndexCrit = currentProfile.MasterFontSettings.Crit.OutlineIndex;


 -- Get the normal outline index for the event.
 local outlineIndex = eventOutlineIndexNormal;

 -- Check if the outline index is inherited from the scroll area outline index.
 if (outlineIndex == 0) then
  outlineIndex = scrollAreaOutlineIndexNormal;
 end

 -- Check if the outline index for the scroll area is inherited from the master outline index.
 if (outlineIndex == 0) then
  outlineIndex = masterOutlineIndexNormal;
 end


 -- Check if the event is a crit.
 if (isCrit) then
  -- Get the crit outline index for the event.
  outlineIndex = eventOutlineIndexCrit;

  -- Check if the outline index is inherited from the scroll area outline index.
  if (outlineIndex == 0) then
   outlineIndex = scrollAreaOutlineIndexCrit;
  end

  -- Check if the outline index for the scroll area is inherited from the master outline index.
  if (outlineIndex == 0) then
   outlineIndex = masterOutlineIndexCrit;
  end
 end


 -- Return the outline style.
 return MikSBT.AVAILABLE_OUTLINES[outlineIndex].Style;
end


-- **********************************************************************************
-- This function substitutes the codes in the unformatted text of the passed
-- animation event with the values in its fields.
-- **********************************************************************************
function MikSBT.FormatEventText(animationEvent)
 local outputString = animationEvent.UnformattedText;

 -- Check if there is no unformatted text and bail.
 if (outputString == nil) then
  return;
 end

 -- Substitute amount.
 if (animationEvent.Amount ~= nil) then
  -- Check if there is overheal info 
  if (animationEvent.OverhealAmount ~= nil and currentProfile.ShowOverheals) then
   local overhealTrailer = string.gsub(MikSBT.MSG_OVERHEAL_TRAILER, "%%d", animationEvent.OverhealAmount);
   outputString = string.gsub(outputString, "%%a", (animationEvent.Amount - animationEvent.OverhealAmount) .. overhealTrailer);
  else 
   outputString = string.gsub(outputString, "%%a", animationEvent.Amount);
  end
 end

 -- Substitute enemy/player names.
 if (animationEvent.Name ~= nil) then
  outputString = string.gsub(outputString, "%%n", animationEvent.Name);
 end

 -- Substitute effect names.
 if (animationEvent.EffectName ~= nil) then
  outputString = string.gsub(outputString, "%%s", animationEvent.EffectName);
 end

 -- Substitute damage types.
 if (animationEvent.DamageType ~= nil) then
  outputString = string.gsub(outputString, "%%t", animationEvent.DamageType);
 end

 -- Check if there is partial effect text and the show partial effects flag is on and append it.
 if (animationEvent.PartialEffectText ~= nil and currentProfile.ShowPartialEffects) then
  outputString = outputString .. animationEvent.PartialEffectText;
 end


 -- Check if there is merge trailer text and append it.
 if (animationEvent.MergeTrailerText ~= nil) then
  outputString = outputString .. animationEvent.MergeTrailerText;
 end

 -- Set the event's output text.
 animationEvent.Text = outputString;
end



-------------------------------------------------------------------------------------
-- Profile functions.
-------------------------------------------------------------------------------------

-- **********************************************************************************
-- Updates profiles created with older versions.
-- **********************************************************************************
function MikSBT.UpdateProfiles()
 -- Loop through all the profiles.
 for _, profile in MikSBT_Save.Profiles do
  -- Check if the profile was created prior to version 2.0.
  if (profile.CreationVersion < 2.0) then
   -- Update the profile's structure.
   profile.ShowOverheals = MikSBT.DEFAULT_CONFIG.ShowOverheals;
   profile.ShowGameDamage = MikSBT.DEFAULT_CONFIG.ShowGameDamage;
   profile.UseStickyCrits = MikSBT.DEFAULT_CONFIG.UseStickyCrits;

   profile.EventSettings["MSBT_EVENTTYPE_INCOMING_SPELL_DOT"] = MikSBT.CopyTable(MikSBT.DEFAULT_CONFIG.EventSettings["MSBT_EVENTTYPE_INCOMING_SPELL_DOT"]);
   profile.EventSettings["MSBT_EVENTTYPE_INCOMING_HOT"] = MikSBT.CopyTable(MikSBT.DEFAULT_CONFIG.EventSettings["MSBT_EVENTTYPE_INCOMING_HOT"]);
   profile.EventSettings["MSBT_EVENTTYPE_OUTGOING_SPELL_DOT"] = MikSBT.CopyTable(MikSBT.DEFAULT_CONFIG.EventSettings["MSBT_EVENTTYPE_OUTGOING_SPELL_DOT"]);
   profile.EventSettings["MSBT_EVENTTYPE_OUTGOING_HOT"] = MikSBT.CopyTable(MikSBT.DEFAULT_CONFIG.EventSettings["MSBT_EVENTTYPE_OUTGOING_HOT"]);

   -- Update the creation version to the current version.
   profile.CreationVersion = 2.0;
  end

  -- Check if the profile was created prior to version 2.1.
  if (profile.CreationVersion < 2.1) then
   -- Update the profile's structure.

   -- Loop through all of the event settings.
   for eventTypeName, eventType in profile.EventSettings do
    -- Check for the event types that can't be crits and remove the crit settings table.
    if (string.find(eventTypeName, "MISS") or string.find(eventTypeName, "DODGE") or string.find(eventTypeName, "PARRY") or
        string.find(eventTypeName, "BLOCK") or string.find(eventTypeName, "ABSORB") or string.find(eventTypeName, "IMMUNE") or
        string.find(eventTypeName, "DOT") or string.find(eventTypeName, "RESIST") or string.find(eventTypeName, "HOT") or
        string.find(eventTypeName, "ENVIRONMENTAL") or string.find(eventTypeName, "EVADE") or string.find(eventTypeName, "NOTIFICATION")) then

     -- Remove the crit settings table.
     eventType.FontSettings.Crit = nil;
    end

    -- Check if the event type is a notification and add the IsSticky flag.
    if (string.find(eventTypeName, "NOTIFICATION")) then
     eventType.IsSticky = MikSBT.DEFAULT_CONFIG.EventSettings[eventTypeName].IsSticky;
    end

    -- Add event settings for experience gains.
    profile.EventSettings["MSBT_EVENTTYPE_NOTIFICATION_EXPERIENCE_GAIN"] = MikSBT.CopyTable(MikSBT.DEFAULT_CONFIG.EventSettings["MSBT_EVENTTYPE_NOTIFICATION_EXPERIENCE_GAIN"]);
   end

   -- Update the creation version to the current version.
   profile.CreationVersion = 2.1;
  end
 end
end


-- **********************************************************************************
-- Returns whether or not the passed profile exists.
-- **********************************************************************************
function MikSBT.ProfileExists(profileName)
 local profileFound = false;

 -- Loop through all of the profiles looking for the passed one.
 for pName in MikSBT_Save.Profiles do
  if (pName == profileName) then
   profileFound = true;   
  end
 end

 -- Return whether or not the profile exists.
 return profileFound;
end


-- **********************************************************************************
-- Resets the passed profile to its defaults.
-- **********************************************************************************
function MikSBT.ResetProfile(profileName, showOutput)
 -- Get whether or not the profile exists.
 local profileExists = MikSBT.ProfileExists(profileName);

 -- Check if the profile exists.
 if (profileExists) then
  -- Reset the profile.
  MikSBT_Save.Profiles[profileName] = MikSBT.CopyTable(MikSBT.DEFAULT_CONFIG);

  -- Check if it's the current profile being reset.
  if (profileName == MikSBT_Save.CurrentProfile) then
   -- Reset the current profile pointer.
   currentProfile = MikSBT_Save.Profiles[MikSBT_Save.CurrentProfile];

   -- Reassociate the display settings if the scroll areas have been setup.
   if (scrollAreas.Incoming ~= nil) then
    -- Setup the display settings.
    scrollAreas.Incoming.DisplaySettings = currentProfile.DisplaySettings.Incoming;
    scrollAreas.Outgoing.DisplaySettings = currentProfile.DisplaySettings.Outgoing;
    scrollAreas.Notification.DisplaySettings = currentProfile.DisplaySettings.Notification;
   end

   -- Reposition the animation frames.
   MikSBT.RepositionAnimationFrames();
  end

  -- Check if the output text is to be shown.
  if (showOutput) then
   -- Print the profile reset string.
   MikSBT.Print(profileName .. " " .. MikSBT.MSG_PROFILE_RESET, 0, 1, 0);
  end
 end 
end


-- **********************************************************************************
-- Selects the passed profile.
-- **********************************************************************************
function MikSBT.SelectProfile(profileName)
 -- Get whether or not the profile exists.
 local profileExists = MikSBT.ProfileExists(profileName);

 -- Check if the profile exists.
 if (profileExists) then
  -- Set the current profile name to the one being selected.
  MikSBT_Save.CurrentProfile = profileName;

  -- Reset the current profile pointer.
  currentProfile = MikSBT_Save.Profiles[MikSBT_Save.CurrentProfile];

  -- Reassociate the display settings if the scroll areas have been setup.
  if (scrollAreas.Incoming ~= nil) then
   -- Setup the display settings.
   scrollAreas.Incoming.DisplaySettings = currentProfile.DisplaySettings.Incoming;
   scrollAreas.Outgoing.DisplaySettings = currentProfile.DisplaySettings.Outgoing;
   scrollAreas.Notification.DisplaySettings = currentProfile.DisplaySettings.Notification;
  end
 end
end


-- **********************************************************************************
-- Creates a profile for the passed name.
-- **********************************************************************************
function MikSBT.CreateProfile(profileName)
 -- Get whether or not the profile exists.
 local profileExists = MikSBT.ProfileExists(profileName);

 -- Make sure a profile name was passed and it doesn't already exist.
 if (not profileExists and profileName ~= nil and profileName ~= "") then
  -- Create an array for the profile.
  MikSBT_Save.Profiles[profileName] = {};
  
  -- Reset the profile.
  MikSBT.ResetProfile(profileName);
 end
end


-- **********************************************************************************
-- Deletes the passed profile.
-- **********************************************************************************
function MikSBT.DeleteProfile(profileName)
 -- Check if the passed profile is the default one.
 if (profileName == MikSBT.DEFAULT_PROFILE_NAME) then
  return;
 end

 -- Get whether or not the profile exists.
 local profileExists = MikSBT.ProfileExists(profileName);

 -- Check if the profile exists.
 if (profileExists) then
  -- Check if the profile being deleted is the current one.
  if (profileName == MikSBT_Save.CurrentProfile) then
   -- Select the default profile.
   MikSBT.SelectProfile(MikSBT.DEFAULT_PROFILE_NAME);
  end

  -- Delete the profile.
  MikSBT_Save.Profiles[profileName] = nil;
 end
end


-------------------------------------------------------------------------------------
-- Option functions.
-------------------------------------------------------------------------------------

-- **********************************************************************************
-- Set the user disabled option
-- **********************************************************************************
function MikSBT.SetOptionUserDisabled(isDisabled)
 MikSBT_Save.UserDisabled = isDisabled;

 -- Check if the mod is being set to disabled.
 if (isDisabled) then
  -- Unregister all events.
  MikSBT.UnregisterEvents();
  MikCEH.UnregisterEvents();

  -- Turn on the game's display of outgoing damage.
  SetCVar("CombatDamage", 1);
 else
  -- Reregister events.
  MikSBT.RegisterEvents();
  MikCEH.RegisterEvents();

  -- Turn the game damage display on or off depending on the user setting.
  MikSBT.SetOptionShowGameDamage(currentProfile.ShowGameDamage);

  -- Make sure that periodic spell effects are reported.
  SetCVar("CombatLogPeriodicSpells", 1);
 end
end


-- **********************************************************************************
-- Set the game damage disabled option
-- **********************************************************************************
function MikSBT.SetOptionShowGameDamage(showDamage)
 currentProfile.ShowGameDamage = showDamage;

 -- Check if the game damage is being set to on.
 if (showDamage) then
   -- Turn on the game damage display.
   SetCVar("CombatDamage", 1);
 else
   -- Turn off the game damage display.
   SetCVar("CombatDamage", 0);
 end
end


-- **********************************************************************************
-- This function takes a string version of a variable path
-- and returns its value from the current profile.
-- **********************************************************************************
function MikSBT.GetOptionFromVarPath(varPath)
 -- Set the current table to the current profile.
 local currentTable = currentProfile;

 -- Index into each sub table until we reach the actual variable wanted.
 for varName in string.gfind(varPath, "[%w_]+") do
  if (type(currentTable[varName]) == "table") then
   currentTable = currentTable[varName];
  else
   return currentTable[varName];
  end
 end

 return nil;
end


-- **********************************************************************************
-- This function takes a string version of a variable path
-- and sets its value in the current profile to the passed value.
-- **********************************************************************************
function MikSBT.SetOptionFromVarPath(varPath, newValue)
 -- Sets the current table to the current profile.
 local currentTable = currentProfile;

 -- Index into each sub table until we reach the actual variable wanted and set it
 -- to the passed value.
 for varName in string.gfind(varPath, "[%w_]+") do
  if (type(currentTable[varName]) == "table") then
   currentTable = currentTable[varName]
  else
   currentTable[varName] = newValue;
  end
 end
end


-------------------------------------------------------------------------------------
-- Animation functions.
-------------------------------------------------------------------------------------

-- **********************************************************************************
-- Adds the passed animation event to the animation system for the appropriate scroll
-- area's animation system.
--
-- animationEvent Possible Fields:
--  DisplayType		- The display type of the event.
--  EventType		- The event type.
--  IsCrit			- Whether or not the event is a crit.
--  UnformattedText	- The unformatted text to display.
--  Amount			- The amount of the attack, heal, gain, loss, etc.
--  EffectName		- The name of the spell, ability, buff, debuff, power type, etc.
--  Name			- The name of the enemy/player.
--  DamageType		- The type of damage.
--  PartialEffectText	- Partial effect text to be appended. (partial resists, absorbs, blocks)
--  OverhealAmount	- The amount overhealed.
--  Text			- The formatted text to display.
-- **********************************************************************************
function MikSBT.DispatchAnimationEvent(animationEvent)
 -- Get the correct scroll area display settings.
 local displaySettings;
 if (animationEvent.DisplayType == MikSBT.DISPLAYTYPE_INCOMING) then
  -- Get the correct data for incoming events.
  displaySettings = currentProfile.DisplaySettings.Incoming;  
 -- Outgoing frame.
 elseif (animationEvent.DisplayType == MikSBT.DISPLAYTYPE_OUTGOING) then
  -- Get the correct data for outgoing events.
  displaySettings = currentProfile.DisplaySettings.Outgoing;  
 elseif (animationEvent.DisplayType == MikSBT.DISPLAYTYPE_NOTIFICATION) then
  -- Get the correct data for notification events.
  displaySettings = currentProfile.DisplaySettings.Notification;  
 end


 -- Check if the event is to be shown.
 if (currentProfile.EventSettings[animationEvent.EventType].Show and displaySettings.Show) then

  -- If it's an incoming or outgoing event add it to the associated scroll area's events array
  -- for potential merging.
  if (animationEvent.DisplayType == MikSBT.DISPLAYTYPE_INCOMING) then
   table.insert(animationMergeData.Incoming.UnmergedEvents, animationEvent);
  elseif (animationEvent.DisplayType == MikSBT.DISPLAYTYPE_OUTGOING) then
   table.insert(animationMergeData.Outgoing.UnmergedEvents, animationEvent);

  -- Add the event the animation system.
  else
   MikSBT.AddAnimation(animationEvent);
  end
 end
end


-- **********************************************************************************
-- Add the event to the appropriate scroll area's animation display array.
-- **********************************************************************************
function MikSBT.AddAnimation(animationEvent)
 -- Format the animation event's output text.
 MikSBT.FormatEventText(animationEvent);

 
 -- Holds the correct scroll area for the event.
 local scrollArea;

 -- Get the correct scroll area for the event.
 if (animationEvent.DisplayType == MikSBT.DISPLAYTYPE_INCOMING) then
  scrollArea = scrollAreas.Incoming;  
 -- Outgoing frame.
 elseif (animationEvent.DisplayType == MikSBT.DISPLAYTYPE_OUTGOING) then
  scrollArea = scrollAreas.Outgoing;
 -- Notification frame.
 elseif (animationEvent.DisplayType == MikSBT.DISPLAYTYPE_NOTIFICATION) then
  scrollArea = scrollAreas.Notification;
 end


 -- Check if the animation event is to be displayed sticky style.
 if (animationEvent.EventType ~= nil and currentProfile.EventSettings[animationEvent.EventType].IsSticky) then
  animationEvent.IsSticky = true;
 end


 -- Get the next available animation display info object for the scroll area.
 local animDisplayInfo = MikSBT.GetNextAnimationDisplayInfo(animationEvent, scrollArea);

 -- Get the starting X and Y positions.
 animDisplayInfo.PositionX, animDisplayInfo.PositionY = MikSBT.GetAnimationStartingPosition(animationEvent, scrollArea)


 -- Check if the animation event has an event type associated with it.
 if (animationEvent.EventType ~= nil) then
  -- Get the font size for the event.
  animDisplayInfo.FontSize = MikSBT.GetFontSize(scrollArea, animationEvent.EventType, animationEvent.IsCrit);

  -- Get the font path for the event.
  animDisplayInfo.FontPath = MikSBT.GetFontPath(scrollArea, animationEvent.EventType, animationEvent.IsCrit);

  -- Get the font outline style for the event.
  animDisplayInfo.FontOutline = MikSBT.GetFontOutline(scrollArea, animationEvent.EventType, animationEvent.IsCrit);

  -- Set the r, g, and b color for the animation display object.
  animDisplayInfo.ColorR = currentProfile.EventSettings[animationEvent.EventType].FontSettings.Color.r;
  animDisplayInfo.ColorG = currentProfile.EventSettings[animationEvent.EventType].FontSettings.Color.g;
  animDisplayInfo.ColorB = currentProfile.EventSettings[animationEvent.EventType].FontSettings.Color.b;

 -- No event type associated with the event so get the display info from the animation event itself.
 else
  -- Get the font size for the event.
  animDisplayInfo.FontSize = animationEvent.FontSize;

  -- Get the font path for the event.
  animDisplayInfo.FontPath = animationEvent.FontPath;

  -- Get the font outline style for the event.
  animDisplayInfo.FontOutline = animationEvent.FontOutline;

  -- Set the r, g, and b color for the animation display object.
  animDisplayInfo.ColorR = animationEvent.ColorR;
  animDisplayInfo.ColorG = animationEvent.ColorG;
  animDisplayInfo.ColorB = animationEvent.ColorB;
 end


 -- Set the animation display alpha, anchor point, crit flag, and sticky flag.
 animDisplayInfo.Alpha = 0;
 animDisplayInfo.AnchorPoint = MikSBT.AVAILABLE_TEXT_ALIGNS[scrollArea.DisplaySettings.FontSettings.TextAlignIndex].AnchorPoint;
 animDisplayInfo.IsCrit = animationEvent.IsCrit;
 animDisplayInfo.IsSticky = animationEvent.IsSticky;

 -- Set the font object properties.
 animDisplayInfo.FontObject:ClearAllPoints();
 animDisplayInfo.FontObject:SetFont(animDisplayInfo.FontPath, animDisplayInfo.FontSize, animDisplayInfo.FontOutline);
 animDisplayInfo.FontObject:SetTextColor(animDisplayInfo.ColorR, animDisplayInfo.ColorG, animDisplayInfo.ColorB);
 animDisplayInfo.FontObject:SetText(animationEvent.Text);
 animDisplayInfo.FontObject:SetAlpha(animDisplayInfo.Alpha);
 animDisplayInfo.FontObject:SetPoint(animDisplayInfo.AnchorPoint, animDisplayInfo.PositionX, animDisplayInfo.PositionY);

 -- Set the last update to the current time.
 animDisplayInfo.LastUpdated = GetTime();

 -- Set the active flag.
 animDisplayInfo.IsActive = true;


 -- Check if the scroll area's display frame is not visible and make it visible so the OnUpdate events start firing.
 -- This is done to keep the number of OnUpdate events down to a minimum for better performance.
 if (not getglobal(scrollArea.AssociatedFrame):IsVisible()) then
  getglobal(scrollArea.AssociatedFrame):Show();
 end

 -- Check if it's a sticky or a crit and sticky crits are enabled.
 if (animDisplayInfo.IsSticky or (animDisplayInfo.IsCrit and currentProfile.UseStickyCrits)) then
  -- Make sure the stickies don't overlap.
  MikSBT.RepositionStickyAnimDisplayInfo(scrollArea);
 else
  -- Make sure there are no overlaps in the scrolling text.
  MikSBT.RepositionAnimDisplayInfo(scrollArea);
 end

 -- Reclaim the animation event table to the events recycler.
 eventsRecycler:ReclaimTable(animationEvent);
end


-- **********************************************************************************
-- This function gets the next animation display info object for the correct
-- scroll area.
-- **********************************************************************************
function MikSBT.GetNextAnimationDisplayInfo(animationEvent, scrollArea)
 -- Holds the next animation display info object.
 local animDisplayInfo;

 -- Check if it's a sticky or it's a crit and sticky crits are enabled.
 if (animationEvent.IsSticky or (animationEvent.IsCrit and currentProfile.UseStickyCrits)) then
  -- Get the next available animation display object for stickies.
  -- This is done so that stickies use the last set of font strings to avoid normal hits
  -- overlapping them.
  animDisplayInfo = scrollArea.AnimationDisplayArray[scrollArea.NextSticky];

  -- Increment to the next sticky animation display object.
  scrollArea.NextSticky = scrollArea.NextSticky + 1;
 
  -- Loop back around to the first sticky animation display object.
  if (scrollArea.NextSticky > NUM_FONT_STRINGS) then
   scrollArea.NextSticky = scrollArea.FirstSticky;
  end

 -- Not a sticky.
 else
  -- Get the next available animation display object.
  animDisplayInfo = scrollArea.AnimationDisplayArray[scrollArea.NextAnimationDisplay];

  -- Increment to the next animation display object.
  scrollArea.NextAnimationDisplay = scrollArea.NextAnimationDisplay + 1;
 
  -- Loop back around to the first animation display object.
  if (scrollArea.NextAnimationDisplay > (scrollArea.FirstSticky - 1)) then
   scrollArea.NextAnimationDisplay = 1;
  end
 end

 -- Return the animation display info object.
 return animDisplayInfo;
end


-- **********************************************************************************
-- This function gets the starting animation x and and y position.
-- **********************************************************************************
function MikSBT.GetAnimationStartingPosition(animationEvent, scrollArea)
 -- Holds where to start the animation at.
 local startPositionX, startPositionY;

 -- Check which animation style is selected for the scroll area.
 if (scrollArea.DisplaySettings.AnimationStyle == ANIMATION_STYLE_STRAIGHT or
     scrollArea.DisplaySettings.AnimationStyle == ANIMATION_STYLE_CURVED_LEFT or
     scrollArea.DisplaySettings.AnimationStyle == ANIMATION_STYLE_CURVED_RIGHT) then

  -- Check if the text is scrolling down.
  if (scrollArea.DisplaySettings.ScrollDirection == SCROLL_DIRECTION_DOWN) then
   -- Set the starting X and Y positions.
   startPositionX = 0;
   startPositionY = scrollArea.DisplaySettings.ScrollHeight;
  -- Text is scrolling up.
  else
   -- Set the starting X and Y positions.
   startPositionX = 0;
   startPositionY = 0;
  end

  -- Check if it's a sticky or a crit and sticky crits are enabled.
  if (animationEvent.IsSticky or (animationEvent.IsCrit and currentProfile.UseStickyCrits)) then
   -- Set the sticky to the center.
   startPositionY = ceil(scrollArea.DisplaySettings.ScrollHeight / 2);
  end
 end

 -- Return the x and y starting positions.
 return startPositionX, startPositionY;
end


-- **********************************************************************************
-- This function makes sure there is no overlapping text.
-- **********************************************************************************
function MikSBT.RepositionAnimDisplayInfo(scrollArea)
 -- Construct a table with the non sticky animation display info in the same order the events were added.
 for x = scrollArea.NextAnimationDisplay, scrollArea.FirstSticky-1 do
  local animDisplayInfo = scrollArea.AnimationDisplayArray[x];
  if (animDisplayInfo.IsActive) then
   table.insert(activeNonStickies, animDisplayInfo);
  end
 end

 for x = 1, scrollArea.NextAnimationDisplay-1 do
  local animDisplayInfo = scrollArea.AnimationDisplayArray[x];
  if (animDisplayInfo.IsActive) then
   table.insert(activeNonStickies, animDisplayInfo);
  end
 end

 -- Get the number of active non sticky data objects.
 local numNonStickies = table.getn(activeNonStickies);

 -- Animation style straight, curved left, or curved right.
 if (scrollArea.DisplaySettings.AnimationStyle == ANIMATION_STYLE_STRAIGHT or 
     scrollArea.DisplaySettings.AnimationStyle == ANIMATION_STYLE_CURVED_LEFT or 
     scrollArea.DisplaySettings.AnimationStyle == ANIMATION_STYLE_CURVED_RIGHT) then

  -- Scrolling down.
  if (scrollArea.DisplaySettings.ScrollDirection == SCROLL_DIRECTION_DOWN) then

   -- Check if there is more than 1 active non sticky animation.
   if (numNonStickies > 1) then
    -- Loop backwards through the active non sticky animation display objects and move the ones that are colliding.
    for x = numNonStickies, 2, -1 do
     local bottomPointCurrent = activeNonStickies[x].PositionY;
     local topPointNext = activeNonStickies[x-1].PositionY + activeNonStickies[x-1].FontSize + MIN_VERTICAL_SPACING;
     if (topPointNext > bottomPointCurrent) then
      activeNonStickies[x-1].PositionY = bottomPointCurrent - activeNonStickies[x-1].FontSize - MIN_VERTICAL_SPACING;
     end
    end
   end

  -- Scrolling up.
  else

   -- Check if there is more than 1 active non sticky animation.
   if (numNonStickies > 1) then
    -- Loop backwards through the active non sticky animation display objects and move the ones that are colliding.
    for x = numNonStickies, 2, -1 do
     local topPointCurrent = activeNonStickies[x].PositionY + activeNonStickies[x].FontSize + MIN_VERTICAL_SPACING;
     local bottomPointNext = activeNonStickies[x-1].PositionY;
     if (bottomPointNext < topPointCurrent) then
      activeNonStickies[x-1].PositionY = topPointCurrent;
     end
    end
   end

  end -- Scrolling Direction
 end -- Animation Style

 -- Clear the temp non stickies table.
 for x = 1, numNonStickies do
  table.remove(activeNonStickies, 1);
 end
end


-- **********************************************************************************
-- This function makes sure stickies don't overlap.
-- **********************************************************************************
function MikSBT.RepositionStickyAnimDisplayInfo(scrollArea)
 -- Construct a table with the sticky animation display info in the same order the events were added.
 for x = scrollArea.NextSticky, NUM_FONT_STRINGS do
  local animDisplayInfo = scrollArea.AnimationDisplayArray[x];
  if (animDisplayInfo.IsActive) then
   table.insert(activeStickies, animDisplayInfo);
  end
 end

 for x = scrollArea.FirstSticky, scrollArea.NextSticky-1 do
  local animDisplayInfo = scrollArea.AnimationDisplayArray[x];
  if (animDisplayInfo.IsActive) then
   table.insert(activeStickies, animDisplayInfo);
  end
 end

 -- Get the number of active stickies.
 local numStickies = table.getn(activeStickies);


 -- Animation style straight, curved left, or curved right.
 if (scrollArea.DisplaySettings.AnimationStyle == ANIMATION_STYLE_STRAIGHT or 
     scrollArea.DisplaySettings.AnimationStyle == ANIMATION_STYLE_CURVED_LEFT or 
     scrollArea.DisplaySettings.AnimationStyle == ANIMATION_STYLE_CURVED_RIGHT) then

  -- Scrolling down.
  if (scrollArea.DisplaySettings.ScrollDirection == SCROLL_DIRECTION_DOWN) then

   -- Check if there is more than 1 active sticky animation.
   if (numStickies > 1) then
    -- Get the middle active sticky.
    local middleSticky = floor((numStickies + 1) / 2);

    -- Set the middle sticky to the center of the scroll area.
    activeStickies[middleSticky].PositionY = scrollArea.DisplaySettings.ScrollHeight / 2;

    -- Loop backwards from the middle sticky.
    for x = middleSticky - 1, 1, -1 do
     activeStickies[x].PositionY = activeStickies[x+1].PositionY - activeStickies[x].FontSize - MIN_VERTICAL_SPACING;
    end

    -- Loop forwards from the middle sticky.
    for x = middleSticky + 1, numStickies do
     activeStickies[x].PositionY = activeStickies[x-1].PositionY + activeStickies[x-1].FontSize + MIN_VERTICAL_SPACING;
    end
   end

  -- Scrolling up.
  else

   -- Check if there is more than 1 active sticky animation.
   if (numStickies > 1) then
    -- Get the middle active sticky.
    local middleSticky = ceil((numStickies + 1) / 2);

    -- Set the middle sticky to the center of the scroll area.
    activeStickies[middleSticky].PositionY = scrollArea.DisplaySettings.ScrollHeight / 2;
   
    -- Loop backwards from the middle sticky.
    for x = middleSticky - 1, 1, -1 do
     activeStickies[x].PositionY = activeStickies[x+1].PositionY + activeStickies[x+1].FontSize + MIN_VERTICAL_SPACING;
    end

    -- Loop forwards from the middle sticky.
    for x = middleSticky + 1, numStickies do
     activeStickies[x].PositionY = activeStickies[x-1].PositionY - activeStickies[x].FontSize - MIN_VERTICAL_SPACING;
    end
   end

  end -- Scrolling Direction
 end -- Animation Style

 -- Clear the temp stickies table.
 for x = 1, numStickies do
  table.remove(activeStickies, 1);
 end
end


-- **********************************************************************************
-- Animates the passed animation display data.
-- **********************************************************************************
function MikSBT.DoAnimation(animDisplayInfo, scrollArea)

 -- Get the current and elapsed times since the last update.
 local currentTime = GetTime();
 local elapsedTime = currentTime - animDisplayInfo.LastUpdated;

 -- Check if it's a sticky or a crit and sticky crits are enabled.
 if (animDisplayInfo.IsSticky or (animDisplayInfo.IsCrit and currentProfile.UseStickyCrits)) then
  -- Move the sticky in case it was repositioned.
  animDisplayInfo.FontObject:SetPoint(animDisplayInfo.AnchorPoint, animDisplayInfo.PositionX, animDisplayInfo.PositionY);

  if (elapsedTime <= STICKY_FADE_IN_TIME) then
   local textDelta = animDisplayInfo.FontSize * (1 - (elapsedTime / STICKY_FADE_IN_TIME));
   animDisplayInfo.FontObject:SetTextHeight(animDisplayInfo.FontSize + textDelta);
   animDisplayInfo.FontObject:SetAlpha(1);
   return;
  elseif (elapsedTime <= (STICKY_DISPLAY_TIME + STICKY_FADE_IN_TIME)) then
   animDisplayInfo.FontObject:SetTextHeight(animDisplayInfo.FontSize);
   return;
  elseif (elapsedTime <= (STICKY_DISPLAY_TIME + STICKY_FADE_IN_TIME + STICKY_FADE_OUT_TIME)) then
   local alpha = 1 - ((elapsedTime - STICKY_DISPLAY_TIME - STICKY_FADE_IN_TIME) / STICKY_FADE_OUT_TIME);
   animDisplayInfo.FontObject:SetAlpha(alpha);
   return;
  end

  -- Deactivate the animation.
  animDisplayInfo.FontObject:SetAlpha(0);
  animDisplayInfo.IsActive = false;

  return;
 end


 -- Check if it's time to update.
 if (elapsedTime > ANIMATION_SPEED) then

  -- Holds whether the animation is complete or not.
  local animationComplete = false;

  -- Calculate the animation midpoint.
  local animationMidPoint = scrollArea.DisplaySettings.ScrollHeight / 2;


  -- Check if the animation style is straight
  if (scrollArea.DisplaySettings.AnimationStyle == ANIMATION_STYLE_STRAIGHT) then

   -- Check if the text is scrolling up.
   if (scrollArea.DisplaySettings.ScrollDirection == SCROLL_DIRECTION_UP) then
    -- Increment the Y position by the step.
    animDisplayInfo.PositionY = animDisplayInfo.PositionY + currentProfile.AnimationStep;

    -- Start fading out the text on the upper half.
    if (animDisplayInfo.PositionY >= animationMidPoint) then
     animDisplayInfo.Alpha = 1 - ((animDisplayInfo.PositionY - animationMidPoint) / animationMidPoint);
    else
     animDisplayInfo.Alpha = 1;
    end

    -- Check if the animation is done.
    if (animDisplayInfo.PositionY >= scrollArea.DisplaySettings.ScrollHeight) then
     -- Set the flag.
     animationComplete = true;
    end

   -- Scrolling down.
   else

    -- Decrement the Y position by the step.
    animDisplayInfo.PositionY = animDisplayInfo.PositionY - currentProfile.AnimationStep;

    -- Start fading out the text on the bottom half.
    if (animDisplayInfo.PositionY < animationMidPoint) then
     animDisplayInfo.Alpha = 1 - ((animationMidPoint - animDisplayInfo.PositionY) / animationMidPoint);
    else
     animDisplayInfo.Alpha = 1;
    end

    -- Check if the animation is done.
    if (animDisplayInfo.PositionY <= 0) then
     -- Set the flag.
     animationComplete = true;
    end
   end -- Scrolling Direction



  -- Animation style curved left.
  elseif (scrollArea.DisplaySettings.AnimationStyle == ANIMATION_STYLE_CURVED_LEFT) then

   -- Check if the text is scrolling up.
   if (scrollArea.DisplaySettings.ScrollDirection == SCROLL_DIRECTION_UP) then
    -- Increment the Y position by the step.
    animDisplayInfo.PositionY = animDisplayInfo.PositionY + currentProfile.AnimationStep;

    -- Start fading out the text on the upper half.
    if (animDisplayInfo.PositionY >= animationMidPoint) then
     animDisplayInfo.Alpha = 1 - ((animDisplayInfo.PositionY - animationMidPoint) / animationMidPoint);
    else
     animDisplayInfo.Alpha = 1;
    end

    -- Check if the animation is done.
    if (animDisplayInfo.PositionY >= scrollArea.DisplaySettings.ScrollHeight) then
     -- Set the flag.
     animationComplete = true;
    end

   -- Scrolling down.
   else

    -- Decrement the Y position by the step.
    animDisplayInfo.PositionY = animDisplayInfo.PositionY - currentProfile.AnimationStep;

    -- Start fading out the text on the bottom half.
    if (animDisplayInfo.PositionY < animationMidPoint) then
     animDisplayInfo.Alpha = 1 - ((animationMidPoint - animDisplayInfo.PositionY) / animationMidPoint);
    else
     animDisplayInfo.Alpha = 1;
    end

    -- Check if the animation is done.
    if (animDisplayInfo.PositionY <= 0) then
     -- Set the flag.
     animationComplete = true;
    end
   end -- Scrolling Direction


   -- Calculate the new x position based on equation of a parabola.
   -- Equation of a parabola at vertex 0,0: x = y^2 / 4a
   local parabolaMaxX = ceil((animationMidPoint * animationMidPoint) / (4 * animationMidPoint));
   local parabolaY = animDisplayInfo.PositionY - animationMidPoint;
   local newX = parabolaMaxX - ceil((parabolaY * parabolaY) / (4 * animationMidPoint));

   -- Set the new x position.
   animDisplayInfo.PositionX = newX * -1;


  -- Animation style curved right.
  elseif (scrollArea.DisplaySettings.AnimationStyle == ANIMATION_STYLE_CURVED_RIGHT) then

   -- Check if the text is scrolling up.
   if (scrollArea.DisplaySettings.ScrollDirection == SCROLL_DIRECTION_UP) then
    -- Increment the Y position by the step.
    animDisplayInfo.PositionY = animDisplayInfo.PositionY + currentProfile.AnimationStep;

    -- Start fading out the text on the upper half.
    if (animDisplayInfo.PositionY >= animationMidPoint) then
     animDisplayInfo.Alpha = 1 - ((animDisplayInfo.PositionY - animationMidPoint) / animationMidPoint);
    else
     animDisplayInfo.Alpha = 1;
    end

    -- Check if the animation is done.
    if (animDisplayInfo.PositionY >= scrollArea.DisplaySettings.ScrollHeight) then
     -- Set the flag.
     animationComplete = true;
    end

   -- Scrolling down.
   else

    -- Decrement the Y position by the step.
    animDisplayInfo.PositionY = animDisplayInfo.PositionY - currentProfile.AnimationStep;

    -- Start fading out the text on the bottom half.
    if (animDisplayInfo.PositionY < animationMidPoint) then
     animDisplayInfo.Alpha = 1 - ((animationMidPoint - animDisplayInfo.PositionY) / animationMidPoint);
    else
     animDisplayInfo.Alpha = 1;
    end

    -- Check if the animation is done.
    if (animDisplayInfo.PositionY <= 0) then
     -- Set the flag.
     animationComplete = true;
    end
   end -- Scrolling Direction


   -- Calculate the new x position based on equation of a parabola.
   -- Equation of a parabola at vertex 0,0: x = y^2 / 4a
   local parabolaMaxX = ceil((animationMidPoint * animationMidPoint) / (4 * animationMidPoint));
   local parabolaY = animDisplayInfo.PositionY - animationMidPoint;
   local newX = parabolaMaxX - ceil((parabolaY * parabolaY) / (4 * animationMidPoint));

   -- Set the new x position.
   animDisplayInfo.PositionX = newX;

  end -- Animation Style


  -- Check if the animation is complete.
  if (animationComplete) then
   -- Set the font object's alpha.
   animDisplayInfo.FontObject:SetAlpha(0);

   -- Deactive the animation.
   animDisplayInfo.IsActive = false;
  else
   -- Move the font object and set its alpha.
   animDisplayInfo.FontObject:SetPoint(animDisplayInfo.AnchorPoint, animDisplayInfo.PositionX, animDisplayInfo.PositionY);
   animDisplayInfo.FontObject:SetAlpha(animDisplayInfo.Alpha);
  end

  -- Set the last update to the current time.
  animDisplayInfo.LastUpdated = currentTime;
 end -- Time for update.
end


-------------------------------------------------------------------------------------
-- Command handler functions.
-------------------------------------------------------------------------------------

-- **********************************************************************************
-- Returns the current and remaining parameters from the passed string.
-- **********************************************************************************
function MikSBT.GetNextParameter(paramString)
 local remainingParams;
 local currentParam = paramString;

 -- Look for a space.
 local idx = string.find(paramString, " ");

 if (idx) then
  -- Get the current and remaing parameters.
  currentParam = string.sub(paramString, 1, idx-1);
  remainingParams = string.sub(paramString, idx+1);
 end

 -- Return the current parameter and the remaining ones.
 return currentParam, remainingParams;
end


-- **********************************************************************************
-- Called to handle commands.
-- **********************************************************************************
function MikSBT.CommandHandler(params)
 -- Get the parameter.
 local currentParam;
 local remainingParams;
 currentParam, remainingParams = MikSBT.GetNextParameter(params);

 -- Flag for whether or not to show usage info.
 local showUsage = true;

 -- Make sure there is a current parameter and lower case it.
 if (currentParam) then
  currentParam = string.lower(currentParam);
 end

 -- Look for the recognized parameters.
 if (currentParam == "") then
  -- Check if the load on demand options are not loaded.
  if (not IsAddOnLoaded(MikSBT.MOD_NAME .. "Options")) then
   -- Load the on demand options.
   UIParentLoadAddOn(MikSBT.MOD_NAME .. "Options");
  end

  -- Make sure the on demand options actually loaded.
  if (IsAddOnLoaded(MikSBT.MOD_NAME .. "Options")) then
   -- Make sure the scroll area mover frame is not shown.
   if (not MSBTScrollAreaMoverFrame:IsShown()) then
    -- Show the options interface.
    MSBTFrameOptions:Show();
   end
  end

  -- Don't show the usage info.
  showUsage = false;

 elseif (currentParam == MikSBT.COMMAND_RESET) then
  -- Reset the current profile.
  MikSBT.ResetProfile(MikSBT_Save.CurrentProfile, true);

  -- Don't show the usage info.
  showUsage = false;

 elseif (currentParam == MikSBT.COMMAND_DISABLE) then
  -- Set the user disabled option.
  MikSBT.SetOptionUserDisabled(true);

  -- Output an informative message.
  MikSBT.Print(MikSBT.MSG_DISABLE, 1,1,1);

  -- Don't show the usage info.
  showUsage = false;

 elseif (currentParam == MikSBT.COMMAND_ENABLE) then
  -- Unset the user disabled option.
  MikSBT.SetOptionUserDisabled(false);

  -- Output an informative message.
  MikSBT.Print(MikSBT.MSG_ENABLE, 1,1,1);

  -- Don't show the usage info.
  showUsage = false;

 elseif (currentParam == MikSBT.COMMAND_SHOWVER) then
  -- Output the current version number.
  MikSBT.Print(MikSBT.VERSION_STRING, 1,1,1);

  -- Don't show the usage info.
  showUsage = false;

 elseif (currentParam == MikSBT.COMMAND_STATS) then
  -- Print out stats about the event recycler.
  eventsRecycler:PrintStats();

  -- Don't show the usage info.
  showUsage = false;

 elseif (currentParam == MikSBT.COMMAND_DEBUG) then
  -- Toggle the debug mode option.
  debugMode = not debugMode;

  -- Check if debug mode was enabled or disabled and set the combat event helper debug flag
  -- appropriately then output the appropriate message.
  if (debugMode) then
   MikSBT.Print(MikSBT.MSG_DEBUG_ENABLE, 1,1,1);
  else
   MikSBT.Print(MikSBT.MSG_DEBUG_DISABLE, 1,1,1);
  end  

  -- Don't show the usage info.
  showUsage = false;
 end

 -- Check if the usage information should be shown.
 if (showUsage) then
  MikSBT.PrintUsage();
 end
end


-- **********************************************************************************
-- Prints out the usage information.
-- **********************************************************************************
function MikSBT.PrintUsage()
 -- Loop through all of the entries in the command usage list.
 for _, msg in MikSBT.COMMAND_USAGE do
  MikSBT.Print(msg, 1, 1, 1);
 end
end